\ JobsPuzzle.ex

\ Example problem for use with LOGICA.
\ Copyright  1992, LogicLab.  All rights reserved.

\ This is a paraphrase of the "Jobs puzzle" presented on pp.55-73 of 
\ Wos, Overbeek, Lusk & Boyle, "Automated Reasoning", Prentice-Hall, Inc., 
\ Englewood Cliffs, NJ 07632 (1984).  This puzzle is typical of the 
\ "tableau problems" presented in puzzle magazines.

\ We have four people: Steve, Pete, Roberta and Thelma.  Each of
\ them holds exactly two jobs.  No two persons hold the same job.
\ The eight jobs are: nurse, chef, guard, telephone operator, 
\ police officer, teacher, actor, boxer.
\ In addition, it is known that:
\ The nurse is a man.
\ The husband of the chef is the telephone operator.
\ Roberta is not a boxer.
\ Pete has only a ninth-grade education.
\ Roberta, the chef, and the police officer went golfing together.

\ The problem is to find out who holds which jobs.

\ To the given data we add some implicit information: that women are 
\ not men, that husbands are always men, that certain jobs (nurse, 
\ teacher, police officer) require a college education, that Roberta 
\ and Thelma are women, that Steve and Pete are men, that the actor 
\ is a man (a woman actor would be an actress).

\ This example illustrates Unit-result resolution, list-directed loops,
\ backdemodulation, explicit constraints, operations on lists and
\ clause weighting.
\
Logic: URRes;    \ UR-resolution
\
Predicate: male, female, is_husband_of, is_wife_of, person, can't_be;
Predicate: more_school, job, jobs, has_job, finished, persons;
Constant: nurse, chef, guard, operator, police, teacher, actor, boxer;
Constant: Steve, Pete, Roberta, Thelma;
Function: jobholder, joblist, personlist;
\
Axiom:
\
\ Cast of characters:
person(x) | 
   FOR(x FROM {steve,pete,roberta,thelma});
\
\ Possible jobs:
Job(x) | 
   FOR(x FROM {chef,guard,nurse,operator,police,teacher,actor,boxer});
\
\ About people:
male(x)  | female(x);
~male(x) | ~female(x);
~is_husband_of(x,x1) | male(x);
~is_wife_of(x,x1) | female(x);
\
\
\ Some limitations on who can hold certain jobs
male(x) | can't_be(x,nurse);
male(x) | can't_be(x,actor);
more_school(x,9) | can't_be(x,police);
more_school(x,9) | can't_be(x,nurse);
more_school(x,9) | can't_be(x,teacher);
~has_job(x,chef) | can't_be(x,police);
~has_job(x,police) | can't_be(x,chef);
is_husband_of(x,jobholder(chef)) | can't_be(x,operator);
is_wife_of(x,jobholder(operator)) | can't_be(x,chef);
\
\ Eliminating possibilities: if person x has job x1, then no one else 
\ has job x1.
~has_job(x,x1) | ~person(x2) | can't_be(x2,x1) | unless(x == x2) ; 
   
\
\ If person x doesn't have job x1, remove x1 from x's list of possible jobs.
~can't_be(x,x1) | rewrite(joblist(x,x2),joblist(x,(x1 DEL x2)));
\
\ If person x doesn't have job x1, remove x from x1's list of possible people.
~can't_be(x,x1) | rewrite(personlist(x1,x2),personlist(x1,(x DEL x2)));
\
\ If we have found two jobs for person x, we're done with him(her).
~jobs(joblist(x,x1)) | finished(x) | while(SIZE x1 = 2);
\
\ If person x has only two possible jobs, they are both his (hers).
~jobs(joblist(x,x1)) | has_job(x,x2) | FOR(x2 FROM x1) | while(SIZE x1 = 2);
\
\ If job x1 has only one possible person, then that person has job x1.
~persons(personlist(x1,x)) | has_job(Head x,x1) | while(SIZE x = 1);
\
\ Termination rule.
~finished(steve) | ~finished(pete) | ~finished(roberta) | ~finished(thelma);
\
Assert: 
\
\ Create joblists to start off
~person(x) | 
 jobs(joblist(x,{chef,guard,nurse,operator,police,teacher, actor, boxer}));
\
\ Create personlists to start off
~job(x) | 
 persons(personlist(x,{steve,pete,roberta,thelma}));
\
\ A few facts:
male(steve);
male(pete);
female(roberta);
female(thelma);
can't_be(roberta,boxer);
can't_be(roberta,chef);
can't_be(roberta,police);
~more_school(pete,9);
\
\ Weights: favor any information about who has or has not a given job;
\ postpone handling "persons" and "jobs" clauses until all other 
\ information has been processed.
Pattern:
Weight: 20  finished(x);
Weight: 50  can't_be(x,x1);
Weight: 50  has_job(x,x1);
Weight: 120 persons(x);
Weight: 150 jobs(x);
\
\ Parameters
LimitC: 20;


